#!/usr/bin/python
# -*- coding: utf-8 -*-
#************************************************************
#
#                       Script
# 
# Université de Lorraine, CNRS, Inria, LORIA, Nancy, France
# 
# Script description: run automatically the experiments of paper "IMITATOR 3: Synthesis of timing parameters beyond decidability" (by Étienne André) published at CAV 2021
# 
# File contributors : Étienne André
# Created           : 2021/04/20
# Last modified     : 2021/04/23
#************************************************************


#############################################################
# Modules
#############################################################

import os
import re
import subprocess
import sys



#############################################################
# Constants
#############################################################

#------------------------------------------------------------
# IMITATOR syntax and options
#------------------------------------------------------------

IMITATOR_TAG_STARTRESULT = 'BEGIN CONSTRAINT'
IMITATOR_TAG_ENDRESULT = 'END CONSTRAINT'

IMITATOR_RESULT_EXTENSION = '.res'

#------------------------------------------------------------
# Binary
#------------------------------------------------------------
IMITATOR_BINARY = "./imitator-v3.0.0-amd64"

#------------------------------------------------------------
# Local setting
#------------------------------------------------------------
BENCHMARKS_DIR = 'models/'
TOOLPAPER_PREFIX = 'result-A-'
EXTRA_PREFIX = 'result-B-'



#############################################################
# Very general functions
#############################################################

#------------------------------------------------------------
# Terminate this script abnormally
#------------------------------------------------------------
def fail_with(text):
    print('Fatal error!')
    print(text)
    sys.exit(1)


#------------------------------------------------------------
# Find text between two substrings or fail
#------------------------------------------------------------
def find_text_between_two_substrings(text, substring1, substring2):
	regexp = substring1 + '(.*)' + substring2
	result = re.search(regexp, text, re.DOTALL)
	
	# Check existence of exactly one result
	groups = result.groups()
	if len(groups) <> 1:
		fail_with('Pattern ' + substring1 + ' (.*) ' + substring2 + ' not found!')
	
	# Return it
	return result.group(1)


#------------------------------------------------------------
# Check that a file exists
#------------------------------------------------------------
def check_file_exists(file_name):
	# Check existence of file
	if not os.path.exists(file_name):
		fail_with('Result file %s does not exist' % file_name)
	
	return True



#############################################################
# Functions for this script
#############################################################

#------------------------------------------------------------
# Run a benchmark with IMITATOR and get the result
#------------------------------------------------------------
def run_benchmark(command_txt, output_prefix):
	# Check existence of binary
	if not os.path.exists(IMITATOR_BINARY):
		fail_with('Binary %s does not exist' % IMITATOR_BINARY)
        
	# Run it
	cmd = [IMITATOR_BINARY] + command_txt.split() + ['-output-prefix'] + [output_prefix]
	
	# Print the command
	print(' command : ' + ' '.join(cmd))
	
	# Run!
	# NOTE: `open(os.devnull, 'wb')` to mute everything
	subprocess.call(cmd, stdout=open(os.devnull, 'wb'), stderr=open(os.devnull, 'wb'))


#------------------------------------------------------------
# Get the result of a computation
#------------------------------------------------------------
def get_constraint(result_file_name):
	# Open the file
	with open(result_file_name, "r") as my_file:
		# Get the content
		result_content = my_file.read()
		
		return find_text_between_two_substrings(result_content, IMITATOR_TAG_STARTRESULT, IMITATOR_TAG_ENDRESULT)

#############################################################
# Definition of all benchmarks to run
#############################################################

tool_paper_benchmarks = [
	# 1. Safety
	{'name' : 'Safety', 'property' : '#synth AGnot(loc[researcher] = coffee)', 'command' : BENCHMARKS_DIR + 'researcher.imi ' + BENCHMARKS_DIR + 'researcher-AGnotcoffee.imiprop'}
	,
	# 2. Reachability
	{'name' : 'Reachability (a)', 'property' : '#synth EF(loc[researcher] = coffee)', 'command' : BENCHMARKS_DIR + 'researcher.imi ' + BENCHMARKS_DIR + 'researcher-EFcoffee.imiprop'}
	,
	# 3. Reachability
	{'name' : 'Reachability (b)', 'property' : '#synth EF(loc[researcher] = coffee & nb >= 5)', 'command' : BENCHMARKS_DIR + 'researcher.imi ' + BENCHMARKS_DIR + 'researcher-EF2coffee.imiprop'}
	,
	# 4. Minimal-time reachability
	{'name' : 'Minimal-time reachability', 'property' : '#synth EFtmin(loc[researcher] = finished & nb >= 2)', 'command' : BENCHMARKS_DIR + 'researcher-globaltime.imi ' + BENCHMARKS_DIR + 'researcher-EFtmin.imiprop'}
	,
	# 5. Optimal parameter reachability
	{'name' : 'Optimal parameter reachability', 'property' : '#synth EFpmin(loc[researcher] = finished & nb >= 3, pTotal)', 'command' : BENCHMARKS_DIR + 'researcher.imi ' + BENCHMARKS_DIR + 'researcher-EFpmin.imiprop'}
	,
	# 6. Parametric deadlock freeness WITH depth limit [NOTE: the analysis wouldn't terminate without the `-depth-limit` option, as written in the paper]
	{'name' : 'Parametric deadlock freeness with depth limit', 'property' : '#synth DeadlockFree', 'command' : BENCHMARKS_DIR + 'researcher.imi ' + BENCHMARKS_DIR + 'researcher-deadlockfree.imiprop -depth-limit 40'}
	,
	# 7. Liveness synthesis
	{'name' : 'Liveness synthesis', 'property' : '#synth CycleThrough(loc[researcher] = finished & nb>=3)', 'command' : BENCHMARKS_DIR + 'researcher.imi ' + BENCHMARKS_DIR + 'researcher-cycle-working-coffee.imiprop'}
	,
	# 8. Robustness
	{'name' : 'Robustness', 'property' : '#synth TracePreservation(pTotal = 10, pNeed = 5, pCoffee = 3, MAXBREAK = 3)', 'command' : BENCHMARKS_DIR + 'researcher.imi ' + BENCHMARKS_DIR + 'researcher-IM.imiprop'}
	,
	# 9. Synthesis using patterns
	{'name' : 'Synthesis using patterns', 'property' : '#synth pattern(everytime restart then eventually done within 5)', 'command' : BENCHMARKS_DIR + 'researcher.imi ' + BENCHMARKS_DIR + 'researcher-pattern-eventually.imiprop'}
	,
	# 10. Generation of Fig.1a
	{'name' : 'Generation of Fig.1a', 'property' : '#synth AGnot(loc[researcher] = coffee)', 'command' : BENCHMARKS_DIR + 'researcher.imi ' + BENCHMARKS_DIR + 'researcher-AGnotcoffee.imiprop -draw-statespace full'}
	,
	# 11. Generation of Fig.1b
	{'name' : 'Generation of Fig.1b', 'property' : '#synth pattern(everytime restart then eventually done within 5)', 'command' : BENCHMARKS_DIR + 'researcher-pNeed2-max3.imi ' + BENCHMARKS_DIR + 'researcher-pattern-eventually.imiprop -draw-cart -draw-cart-x-max 6 -draw-cart-y-max 6'}
	,
]



#############################################################
# Part 1: reproduction of the toy example executions in the tool paper
#############################################################

print("\n HELLO! This is IMITATOR 3’s tool paper artifact evaluation")

print("\n\n#############################################################")
print(" Part 1: reproduction of the toy example executions in the tool paper")
print("#############################################################")

benchmark_id = 1

# For each benchmark
for benchmark in tool_paper_benchmarks:
	
	# Write some information
	print("------------------------------------------------------------")
	print('Property ' + str(benchmark_id) + ': ' + benchmark['name'])
	print(benchmark['property'])

	# Build the output prefix
	output_prefix = BENCHMARKS_DIR + TOOLPAPER_PREFIX + str(benchmark_id)
	
	# Run
	run_benchmark(benchmark['command'], output_prefix)

	# Build result file name
	result_file_name = output_prefix + IMITATOR_RESULT_EXTENSION

	# Check that the result is successfully created
	if not check_file_exists(result_file_name):
		print('Error! File `' + result_file_name + '` not found!')

	# Get the result
	result = get_constraint(result_file_name)

	print('Result:\n' + result)

	benchmark_id = benchmark_id + 1



#############################################################
# Part 2: reproduction of additional experiments
#############################################################

print("\n\n#############################################################")
print(" Part 2: reproduction of additional experiments")
print("#############################################################")

industrial_benchmarks = [
	# 1. SPSMALL
	{'name' : 'SPSMALL', 'command' : BENCHMARKS_DIR + 'spsmall.imi ' + BENCHMARKS_DIR + 'spsmall:AGnot.imiprop'}
	,
	# 2. Thales
	{'name' : 'Thales FMTV challenge', 'command' : BENCHMARKS_DIR + 'FMTV-1-A1.imi ' + BENCHMARKS_DIR + 'FMTV-1-A1:EF.imiprop'}
	,
	# 3. Parametric pattern matching and monitoring logs from the automative industry
	{'name' : 'Parametric pattern matching and monitoring logs from the automative industry', 'command' : BENCHMARKS_DIR + 'accel:3000.imi ' + BENCHMARKS_DIR + 'accel:EF.imiprop -no-cumulative-pruning'}
	,
	# 4. Parametric timed opacity analysis
	{'name' : 'Parametric timed opacity analysis', 'command' : BENCHMARKS_DIR + 'category3_vulnerable-selfcomp-param.imi ' + BENCHMARKS_DIR + 'category3_vulnerable.imiprop'}
	,
	# 5. Verification of the Bounded Retransmission Protocol
	{'name' : 'Bounded Retransmission Protocol', 'command' : BENCHMARKS_DIR + 'brp_Channels.imi ' + BENCHMARKS_DIR + 'brp_Channels.imiprop'}
	,
]

benchmark_id = 1

# For each benchmark
for benchmark in industrial_benchmarks:
	
	# Write some information
	print("------------------------------------------------------------")
	print('Benchmark ' + str(benchmark_id) + ': ' + benchmark['name'])

	# Build the output prefix
	output_prefix = BENCHMARKS_DIR + EXTRA_PREFIX + str(benchmark_id)
	
	# Run
	run_benchmark(benchmark['command'], output_prefix)

	# Build result file name
	result_file_name = output_prefix + IMITATOR_RESULT_EXTENSION

	# Check that the result is successfully created
	if not check_file_exists(result_file_name):
		print('Error! File `' + result_file_name + '` not found!')

	# Get the result
	result = get_constraint(result_file_name)

	print('Result:\n' + result)

	benchmark_id = benchmark_id + 1


print("\nDone!")
